from .common import *

def Message(content, close, show, class_names="fixed top-0 px-4 py-4 right-0 mt-2  shadow z-40 rounded bg-light-white dark:bg-dark-black"):
	"""
	Creates a closeable message notification
	
	args:
			text: content of message
			close: close handler
			class_names : Default class names
			show : boolean show or hide message
	"""
	main = "relative" if show else "hidden"
	return Div(main, [
		Div(class_names, [
			Div("relative", [
				content,
				Button("fixed top-0 right-0 dark:text-dark-page1 ", "X", attrs={
				"py-click": close
				})
			])
		])
	])



from zenaura.client.tags.builder import Builder

def Card(
    content,
    attrs, 
    default_class="shadow  w-64 me-2 px-2.5 py-0.5 rounded bg-light-white text-light-gray1 dark:text-dark-page1 dark:bg-dark-gray2", 
    ):
    """
      Displays a a card with content. 
      args:
        content : card content of zenaura components or elements.
        attrs: card attributes.
        default_class: default css class names.
      
    """
    return Builder("div").with_attribute("class", default_class).with_attributes(**attrs).with_children(*content).build()



from .common import *


def Popover(
  content,
  over_content,
  show=False,
  position="bottom-0"
  ):
  """
    Displays a button and popover appears when button is hovered over. 
    args:
      content : popover content
      attrs: card attributes.
      default_class: default css class names for content
      over_content: content that popover appear when overed over
      show: when user hover over over_content content card will be visible, elease will be hidden
      position: position of content to over_content, left-0, right-0, top-0, bottom-0 relative position
  """
  return Div(
      "relative",
    [
        over_content,
      Div(f"absolute z-10  {position}" + (" hidden" if not show else ""), [
        content
      ]),
    ]
    )


from zenaura.client.tags.builder import Builder
from zenaura.client.tags.node import Attribute
from .styles import with_theme_colors, btn_one_class, with_theme_colors_text_no_hover
from typing import List

# Base 
def Image(src, alt, width, height, classname=""):
	"""
	Creates an HTML image element.
	Args:
			src (str): The source URL of the image.
			alt (str): The alternate text for the image.
			width (str): The width of the image.
			height (str): The height of the image.
			classname (str): The class name for the image.
	Returns:
			An HTML img element with the specified attributes.
	"""
	return Builder("img").with_attributes(
		src=src,
		alt=alt,
		width=width,
		height=height,
	).with_attribute("class", classname).build()

def Header2(text, class_name=""):
	"""
	Creates an HTML h2 element.
	Args:
			text (str): The text content of the header.
			class_name (str): The class name for the header.
	Returns:
			An HTML h2 element with the specified attributes and text.
	"""
	return Builder('h2').with_attribute("class", class_name).with_text(text).build()

def Header1(text, class_names):
	"""
	Creates an HTML h1 element.
	Args:
			text (str): The text content of the header.
			class_names (str): The class names for the header.
	Returns:
			An HTML h1 element with the specified attributes and text.
	"""
	return Builder('h1').with_text(text).with_attribute("class", class_names).build()

def Section(children, class_name="intro"):
	"""
	Creates an HTML section element.
	Args:
			children (list): The child elements of the section.
			class_name (str): The class name for the section.
	Returns:
			An HTML section element with the specified attributes and children.
	"""
	section = Builder('section').with_attribute('class', class_name).build()
	section.children = children
	return section

def HR():
	"""
	Creates an HTML hr element.
	Returns:
			An HTML hr element with specified attributes.
	"""
	return Builder("hr").with_attribute('class', "w-full border-b-1 border-light-green dark:border-gray-700 ").build()

def OL(children, attrs):
	"""
	Creates an HTML ordered list (ol) element.
	Args:
			children (list): The child li elements.
			attrs (dict): The attributes for the ol element.
	Returns:
			An HTML ol element with the specified attributes and children.
	"""
	return Builder("ol").with_attributes(**attrs).with_children(*children).build()

def LI(child, attrs):
	"""
	Creates an HTML list item (li) element.
	Args:
			child (element): The child element of the li.
			attrs (dict): The attributes for the li element.
	Returns:
			An HTML li element with the specified attributes and child.
	"""
	return Builder("li").with_attributes(**attrs).with_child(child).build()

def A(child, attrs):
	"""
	Creates an HTML anchor (a) element.
	Args:
			child (element): The child element of the a.
			attrs (dict): The attributes for the a element.
	Returns:
			An HTML a element with the specified attributes and child.
	"""
	return Builder("a").with_attributes(**attrs).with_child(child).build()

def Dialog(children, attrs):
	"""
	Creates an HTML dialog element.
	Args:
			children (list): The child elements of the dialog.
			attrs (dict): The attributes for the dialog element.
	Returns:
			An HTML dialog element with the specified attributes and children.
	"""
	return Builder("dialog").with_children(*children).with_attributes(**attrs).build()

# features menu

def Paragraph(text, class_name=None):
	"""
	Creates an HTML paragraph (p) element.
	Args:
			text (str): The text content of the paragraph.
			class_name (str, optional): The class name for the paragraph.
	Returns:
			An HTML p element with the specified attributes and text.
	"""
	builder = Builder('p').with_text(text)
	if class_name:
			builder = builder.with_attribute('class', class_name)
	return builder.build()

def Div(class_name, children):
	"""
	Creates an HTML div element.
	Args:
			class_name (str): The class name for the div.
			children (list): The child elements of the div.
	Returns:
			An HTML div element with the specified attributes and children.
	"""
	div = Builder('div').with_attribute('class', class_name).build()
	div.children = children
	return div

def Button(class_name, text, onclick_handler=None, name=None, attrs={}):
	"""
	Creates an HTML button element.
	Args:
			class_name (str): The class name for the button.
			text (str): The text content of the button.
			onclick_handler (str, optional): The JavaScript function to call on click.
			name (str, optional): The name of the button.
			attrs (dict): Additional attributes for the button.
	Returns:
			An HTML button element with the specified attributes and text.
	"""
	builder = Builder('button').with_attribute('class', class_name).with_text(text)
	if onclick_handler:
			builder = builder.with_attribute('py-click', onclick_handler)
	if name:
			builder = builder.with_attribute("name", name)
	return builder.with_attributes(**attrs).build()

def ButtonWithAttrsChildren(class_name, attrs, children, onclick_handler=None, name=None):
	"""
	Creates an HTML button element with specified attributes and children.
	Args:
			class_name (str): The class name for the button.
			attrs (dict): Additional attributes for the button.
			children (list): The child elements of the button.
			onclick_handler (str, optional): The JavaScript function to call on click.
			name (str, optional): The name of the button.
	Returns:
			An HTML button element with the specified attributes and children.
	"""
	return Builder("button") \
			.with_attribute("class", class_name) \
			.with_attributes(**attrs).with_children(*children) \
			.with_attribute("py-click", onclick_handler) \
			.build()

def Loader():
	"""
	Creates a loader div element.
	Returns:
			An HTML div element representing a loader.
	"""
	return Div("loader self-center bg-light-white dark:bg-dark-gray1", [
		Div("", [
			Div("", [
			])
		])
	])

# nav 

def NavItemText(href, text, class_names, click=None):
	"""
	Creates a navigation item with text.
	Args:
			href (str): The URL the item links to.
			text (str): The text content of the item.
			class_names (str): The class names for the item.
			click (str, optional): The JavaScript function to call on click.
	Returns:
			An HTML a element with the specified attributes and text.
	"""
	tag = Builder('a') \
		.with_attribute("class", class_names) \
		.with_attribute("href", href) \
		.with_text(text)
	if click:
			tag.with_attribute("py-click", click)
	return tag.build()

def NavItemTextNameFactory(href, text, class_names, click=None):
	"""
	Creates a navigation item with text and a name attribute.
	Args:
			href (str): The URL the item links to.
			text (str): The text content of the item.
			class_names (str): The class names for the item.
			click (str, optional): The JavaScript function to call on click.
	Returns:
			An HTML a element with the specified attributes, text, and name.
	"""
	tag = Builder('a') \
		.with_attribute("class", class_names) \
		.with_attribute("href", href) \
		.with_text(text)
	if click:
		tag.with_attribute("py-click", click)
	name = text.lower()
	tag.with_attribute("name", name)
	return tag.build()

def Link(href, text, class_names, target="_blank"):
	"""
	Creates an HTML anchor (a) element.
	Args:
			href (str): The URL the link points to.
			text (str): The text content of the link.
			class_names (str): The class names for the link.
			target (str): The target attribute for the link.
	Returns:
			An HTML a element with the specified attributes and text.
	"""
	return Builder('a') \
		.with_attribute("class", class_names) \
		.with_attribute("href", href) \
		.with_attribute("target", target) \
		.with_text(text).build()
       
def NavItemIcon(href, img, class_names=""):
	"""
	Creates a navigation item with an icon.
	Args:
			href (str): The URL the item links to.
			img (element): The image element for the icon.
			class_names (str): The class names for the item.
	Returns:
			An HTML a element with the specified attributes and icon.
	"""
	return Builder('a').with_attribute('href', href).with_attribute("class", class_names).with_child(img).build()

def SvgPath(linecap, linejoin, d):
	"""
	Creates an SVG path element.
	Args:
			linecap (str): The stroke-linecap attribute.
			linejoin (str): The stroke-linejoin attribute.
			d (

str): The path data.
	Returns:
			An SVG path element with the specified attributes.
	"""
	return Builder('path') \
		.with_attribute('stroke-linecap', linecap) \
		.with_attribute('stroke-linejoin', linejoin) \
		.with_attribute('d', d) \
		.build()

def Svg(class_name, fill, viewBox, stroke, path, stroke_width=None):
	"""
	Creates an SVG element.
	Args:
			class_name (str): The class name for the SVG.
			fill (str): The fill attribute.
			viewBox (str): The viewBox attribute.
			stroke (str): The stroke attribute.
			path (element): The path element inside the SVG.
			stroke_width (str, optional): The stroke-width attribute.
	Returns:
			An SVG element with the specified attributes and path.
	"""
	svg = Builder('svg') \
		.with_attribute('class', class_name) \
		.with_attributes(
				fill=fill,
				viewBox=viewBox,
				stroke=stroke
		).with_child(
				path
		)
	if stroke_width:
			svg.with_attribute('stroke-width', stroke_width)
	return svg.build()

def Span(class_name, text=None):
	"""
	Creates an HTML span element.
	Args:
			class_name (str): The class name for the span.
			text (str, optional): The text content of the span.
	Returns:
			An HTML span element with the specified attributes and text.
	"""
	span = Builder('span').with_attribute('class', class_name)
	if text:
			span.with_text(text)
	return span.build()


import unittest
import io
from zenaura.client.page import Page 
from .mocks.counter_mocks import Counter
from zenaura.client.hydrator import HydratorCompilerAdapter
from zenaura.server.server import ZenauraServer, template
from zenaura.client.app import App 
from unittest.mock import MagicMock, patch
from zenaura.client.layout import Layout


class TestZenauraServer(unittest.TestCase):

    def setUp(self):
        from zenaura.client.app import HistoryNode, PageHistory, NotFound, App, Route
        from .mocks.counter_mocks import Counter
        from .mocks.browser_mocks import MockDocument, MockWindow
        self.page = Page([Counter([])])
        self.page.id = 'test-page-id'
        self.page_content = '<div>Test Page Content</div>'
        self.compiler_adapter = HydratorCompilerAdapter()
        self.maxDiff = None
        self.middleware = MagicMock()
        self.page = Page([Counter([])])
        self.route = Route("test", "/test", self.page, self.middleware) 
        self.route_home = Route("test3", "/", self.page, self.middleware) 
        self.route2 = Route("test2", "/2", self.page, self.middleware) 
        self.new_route = Route
        self.router = App()
        self.router.add_route(self.route)
        self.router.add_route(self.route_home)
        self.router.add_route(self.route2)

        self.history_node = HistoryNode
        self.page_history = PageHistory
        self.not_found = NotFound()
        self.document = MockDocument()
        self.window = MockWindow()
        self.dom_patcher = patch("zenaura.client.app.document", self.document)
        self.dom_patcher.start()
        self.window_patcher = patch("zenaura.client.app.window", self.window)
        self.window_patcher.start()

        self.router_without_home = App()
        self.router_without_home.add_route(self.route)
        self.router_without_home.add_route(self.route2)
        self.page_with_attrs = Page([Counter([])], {"k" : "test"})
        self.router_page_with_attrs = App()
        self.router_page_with_attrs.add_route(
            Route("test", "/test",self.page_with_attrs, self.middleware )
        )
        
        global compiler_adapter
        compiler_adapter = self.compiler_adapter  # Mocking the global instance
        
        self.template_content = template

    def tearDown(self):
        self.dom_patcher.stop()
        self.window_patcher.stop()

    def test_hydrate_page(self):
        result = ZenauraServer.hydrate_page(self.page)
        self.assertEqual(result, self.template_content(self.compiler_adapter.hyd_comp_compile_page(self.page), title="zenaura", meta_description="this app created with zenaura", icon="./public/favicon.ico", pydide="https://pyscript.net/releases/2024.1.1/core.js"))

    @patch("builtins.open", new_callable=unittest.mock.mock_open)
    def test_hydrate_app_home_defined(self, mock_open):
        result = ZenauraServer.hydrate_app(self.router)
        routes = self.router.routes 
        pages = []
        for _, route in routes.items():
            page, _, _, _ = route
            pages.append(page.id)
        # / path is not hidden
        self.assertIn(
            f'<div data-zenaura="{pages[0]}">',
            result, 
        )
        # rest are hidden
        self.assertIn(
            f'<div hidden data-zenaura="{pages[1]}">',
            result, 
        )
        self.assertIn(
            f'<div hidden data-zenaura="{pages[2]}">',
            result, 
        )
    
    @patch("builtins.open", new_callable=unittest.mock.mock_open)
    def test_hydrate_app_home_not_defined_first_in_stack_shown(self, mock_open):        
        result = ZenauraServer.hydrate_app(self.router_without_home)
        
        routes = self.router_without_home.routes 

        pages = []
        for _, route in routes.items():
            page, _, _, _ = route
            pages.append(page.id)
        # / path is not hidden
        self.assertIn(
            f'<div hidden data-zenaura="{pages[0]}">',
            result, 
        )
        # rest are hidden
        self.assertIn(
            f'<div hidden data-zenaura="{pages[1]}">',
            result, 
        )


    @patch("builtins.open", new_callable=unittest.mock.mock_open)
    def test_hydrate_app_pages_with_attrs(self, mock_open):        
        result = ZenauraServer.hydrate_app(self.router_page_with_attrs)
        
        routes = self.router_page_with_attrs.routes 

        pages = []
        for _, route in routes.items():
            page, _, _, _ = route
            pages.append(page.id)
        # / path is not hidden
        self.assertIn(
            f'<div k="test" data-zenaura="{pages[0]}">',
            result, 
        )

    @patch("builtins.open", new_callable=unittest.mock.mock_open)
    def test_hydrate_app_layout_pages_with_attrs(self, mock_open):
        layout = Layout([Counter([])], self.router_page_with_attrs.routes,[Counter([]) ] )
        result = ZenauraServer.hydrate_app_layout(layout)
        routes = layout.routes
        pages = []
        for _, route in routes.items():
            page, _, _, _ = route
            pages.append(page.id)

        # Page with attributes is not hidden
        self.assertIn(f'<div k="test" data-zenaura="{pages[0]}">', result)
        self.assertEqual(result.count(f'data-zenaura="{Counter([]).id}"') , 3)

    
if __name__ == '__main__':
    unittest.main()



from .common import *

def Menu(main_btn : Button, children: List[ButtonWithAttrsChildren], show : bool) -> "Menu":
	"""
		Display a menu to the user - triggered by py-click

		args : 
			main_btn -> button used to toggle dropdown menu on click.
			children -> menu children 
			show -> used to toggle menu visibility
		return : 
			Togglable menu with options
	"""

	menu = Div(
		"absolute right-0 z-20 w-56 py-2 mt-2 overflow-hidden bg-white rounded-md shadow-xl dark:bg-dark-gray1" + (" hidden" if not show else ""),
		children
	)

	return Div("flex justify-center", [
		Div("relative inline-block mb-20", [
			main_btn, 
			menu

		])
	])



from .common import *


def Modal(content, close_btn, show_modal=False, class_names="shadow z-40 rounded bg-light-white text-light-gray1 dark:text-dark-page1 dark:bg-dark-gray2"):
  """
  Creates a modal component with given content.
  
  args:
      content: Content to be displayed inside the modal, typically a Component.
      show_modal: Boolean to control the visibility of the modal.
      close_btn: Button to close modal when it's open
      class_names : Default class names
  """
  attrs = {"open": "", "class": class_names} if show_modal else {}
  modal_content = Dialog([
    content,
    close_btn,
  ], attrs)
  
  return modal_content


import functools
import asyncio 
from zenaura.client.dispatcher import dispatcher
from zenaura.client.dom import zenaura_dom



def mutator(coroutine):
    """
    Decorator to automatically render the DOM after the decorated coroutine executes.

    This decorator is useful for functions that modify the state of the application,
    triggering a re-rendering of the DOM to reflect the changes.

    Args:
        coroutine: The function to be decorated.

    Returns:
        wrapper_func: A wrapper function that executes the original function and then renders the DOM.

    Raises:
        TypeError: If `func` is not a callable.
    """

    if not callable(coroutine):
        raise TypeError("mutator can only decorate callables")
    
    if not asyncio.iscoroutine(coroutine) and not asyncio.iscoroutinefunction(coroutine):
        raise TypeError(f"'{coroutine}' is not a valid aysncio coroutine. use @mutates to decorate regular python callbacks'.")


    @functools.wraps(coroutine)
    async def wrapper_func(component, *args, **kwargs):
        """
        Wrapper function that executes the original function and then renders the DOM.

        Args:
            component: The instance of the class.
            *args: Positional arguments for the original function.
            **kwargs: Keyword arguments for the original function.

        Returns:
            None
        """
        dispatcher.dispatch(coroutine, component, *args, kwargs)
        dispatcher.dispatch(zenaura_dom.render, component)

    return wrapper_func


def mutates(func):
    """
    Decorator to automatically render the DOM after the decorated function executes.

    This decorator is useful for functions that modify the state of the application,
    triggering a re-rendering of the DOM to reflect the changes.

    Args:
        coroutine: The function to be decorated.

    Returns:
        wrapper_func: A wrapper function that executes the original function and then renders the DOM.

    Raises:
        TypeError: If `func` is not a callable.
    """

    if not callable(func):
        raise TypeError("mutator can only decorate functions")
    
    if asyncio.iscoroutine(func) or asyncio.iscoroutinefunction(func):
        raise TypeError(f"'{func}' is an a aysncio coroutine. use @mutator to decorate coroutines'.")


    @functools.wraps(func)
    def wrapper_func(component, *args, **kwargs):
        """
        Wrapper function that executes the original function and then renders the DOM.

        Args:
            component: The instance of the class.
            *args: Positional arguments for the original function.
            **kwargs: Keyword arguments for the original function.

        Returns:
            None
        """
        func(component, *args, kwargs)
        dispatcher.dispatch(zenaura_dom.render, component)

    return wrapper_func



from .common import *


def Select(
  label_text,
  attrs,
  options, 
  default_input_class="w-full p-2 border border-gray-300 rounded bg-light-white text-light-gray1 dark:text-dark-page1 hover:bg-light-green dark:bg-dark-gray2 dark:hover:bg-dark-gray1", 
  default_label_class="block mb-2 text-light-gray1 dark:text-dark-page1",
  default_wrapper_class="p-4",
  ):
  """
    Displays a form input field. 
    args:
      attrs: python dictionary trasfrom into html attributes.
      label: label_text : input label text.
      default_input_class : default input css class names
      default_label_class : default label css class names
      default_wrapper_class : default input wrapper css class names
      options: select options List[Option]
    
  """
  return Div(default_wrapper_class, [
    Builder("label").with_attribute("class", default_label_class).with_text(label_text).build(),
    Builder('select').with_attribute("class", default_input_class).with_attributes(
      **attrs
    ).with_children(*options).build(),
  ])

def Option(
  label,
  attrs,
  default_input_class="w-full p-2 border border-gray-300 rounded bg-light-white text-light-gray1 dark:text-dark-page1 hover:bg-light-green dark:bg-dark-gray2 dark:hover:bg-dark-gray1", 
  ):
  """
    Displays a form input field. 
    args:
      attrs: python dictionary trasfrom into html attributes.
      default_input_class : default input css class names
      default_label_class : default label css class names
      value : text value of option      
  """
  return Builder('option').with_attribute("class", default_input_class).with_attributes(
    **attrs
  ).with_text(label).build()



import unittest
from unittest.mock import patch, AsyncMock, MagicMock, AsyncMock
from zenaura.client.dispatcher import AsyncDispatcher
from tests.mocks.browser_mocks import MockDocument, MockWindow, create_proxy, MockElement

class TestAsyncDispatcher(unittest.TestCase):
    def setUp(self):
        self.dispatcher = AsyncDispatcher()

    def tearDown(self):
        if self.dispatcher.loop and not self.dispatcher.loop.is_closed():
            self.dispatcher.loop.close()
    
    @patch("zenaura.web.utils.document")
    def test_dispatch_runs_coroutine(self, document):
        """Test that dispatch runs a coroutine."""
        async def sample_coro(arg1, arg2):
            return arg1 + arg2
        
        document.ready_state = "interactive"

        coro_mock = AsyncMock(side_effect=sample_coro)

        self.dispatcher.dispatch(coro_mock, 2, 3)

        coro_mock.assert_awaited_with(2, 3)

    @patch("zenaura.web.utils.document.getElementById")
    def test_bind_event_on_element(self, getElementById):
        """Test bind method binds events to elements."""

        async def test_callback(event):
            pass

        self.dispatcher.bind("test_id", "click", test_callback)

        getElementById.assert_called_with("test_id")

        
    @patch("zenaura.web.utils.document.addEventListener")
    def test_bind_to_window(self, addEventListener):
        """Test bind attaches an event to the window."""
        
        async def test_callback(event):
            pass

        self.dispatcher.bind("document", "load", test_callback)
        addEventListener.assert_called_once()


    @patch("zenaura.web.utils.window.addEventListener")
    def test_bind_to_window(self, addEventListener):
        """Test bind attaches an event to the window."""
        
        async def test_callback(event):
            pass

        self.dispatcher.bind("window", "resize", test_callback)
        addEventListener.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import asyncio 
from zenaura.web.utils import document, window, in_browser, create_proxy
import nest_asyncio
nest_asyncio.apply()
class AsyncDispatcher:
    def __init__(self):
        self.loop = None

    def dispatch(self, coro_func, *args, **kwargs):
        """
        Wrap asyncio.run to schedule and run the given coroutine.
        
        :param coro_func: The coroutine function to execute.
        :param args: Positional arguments for the coroutine.
        :param kwargs: Keyword arguments for the coroutine.
        """
        if not self.loop or self.loop.is_closed():
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            
        self.loop.run_until_complete(coro_func(*args, **kwargs))

    def bind(self, id, event, coroutine):
        """
        Subscribe an event to an element, window, or document and dispatch a sync or async callback.
        Ensures the provided function has the correct signature.

        :param id: ID attached to the component, or 'window'/'document' for global bindings.
        :param event: Event name (e.g., "click", "resize", "keydown").
        :param coroutine: Callback coroutine function.
        :return: None if binding succeeds; raises an exception otherwise.
        """

        if not callable(coroutine):
            raise TypeError("The callback must be callable.")

        if not asyncio.iscoroutine(coroutine) and not asyncio.iscoroutinefunction(coroutine):
            raise TypeError(f"'{coroutine}' is not a valid asyncio coroutine. The callback must be 'async def'.")

        target = None

        # Determine the target based on id
        if id == "window":
            target = window
            print("yes window", target)
        elif id == "document":
            target = document
        else:
            target = document.getElementById(id)

        # if user not in development server and target not found
        # throw an error 
        if not target :
            if not in_browser:
              raise ValueError(f"Element with id '{id}' not found. Cannot bind '{coroutine.__name__}' to '{event}'.")

        # Bind the event listener
        try:
            target.addEventListener(event, create_proxy(lambda e: self.dispatch(coroutine, e)))
        except Exception as e:
            raise RuntimeError(f"Failed to bind '{coroutine.__name__}' to '{event}' on target '{id}': {e}")


dispatcher = AsyncDispatcher()


from zenaura.client.tags.builder import Builder


def Form(
    attrs,
    fields,
    on_submit,
    default_form_styles="w-full p-2 border border-gray-300 rounded bg-light-white text-light-gray1 dark:text-dark-page1 hover:bg-light-green dark:bg-dark-gray2 dark:hover:bg-dark-gray1", 
    ):
    """
      Displays a form input field. 
      args:
        attrs: python dictionary trasfrom into html attributes.
        default_form_styles : default input css class names for form
        fields: list of zenaura input fields or custom elements
      
    """
    return Builder("form").with_attribute("class", default_form_styles).with_attributes(**attrs).with_children(*fields).with_attribute("py-submit", on_submit).build()



from zenaura.client.tags.builder import Builder
from zenaura.web.utils import document, to_js
def ChartThis(config, chart_id, *args, **kwargs):
  """
    Attaches a chart to canvas.

    Args: 
      config - ChartJS config https://www.chartjs.org/docs/latest/configuration/
      chart_id - unique id for canvas <canvas id=chart_id> to attach chartjs graph to.
    returns:
      Chart
  """
  from js import Chart, Object
  # clean up chart if chart exists 
  chart_exists = Chart.getChart(chart_id)
  if chart_exists:
    chart_exists.destroy()
  
  # use pyodide to transfrom python dict into js object
  obj = to_js(config, depth=-1,  
                  pyproxies=None,
      create_pyproxies=False,
      dict_converter=Object.fromEntries)
  
  # attach graph
  ctx = document.getElementById(chart_id).getContext('2d')
  Chart.new(ctx, obj)

  # return Chart for extendibility
  return Chart

def Canvas(id, attrs={}, childrens=[]):
  """
    canvas html tag for use with chartjs

    args: 
      id - unique chart id
      attrs - html attributes
      children - children within canvas tag if needed
  """
  return  Builder("canvas").with_attribute("id", id).build()


btn_one_class = """
inline-flex items-center justify-center 
whitespace-nowrap text-sm font-medium 
transition-colors focus-visible:outline-none 
focus-visible:ring-1 focus-visible:ring-ring 
disabled:pointer-events-none disabled:opacity-50 
text-primary-foreground shadow 
h-9 px-4 py-2 rounded-[6px]
m-1 bg-light-gray1 text-light-white
hover:bg-light-green
dark:text-dark-page1
dark:bg-dark-black
dark:hover:bg-dark-gray2
"""

btn_two_class ="""
inline-flex items-center justify-center whitespace-nowrap 
text-sm font-medium transition-colors focus-visible:outline-none 
focus-visible:ring-1 focus-visible:ring-ring 
disabled:pointer-events-none disabled:opacity-50 
border border-input bg-background shadow-sm
h-9 px-4 py-2 rounded-[6px]
m-1 bg-light-white text-light-gray1
hover:text-light-green
dark:text-dark-black
dark:hover:bg-dark-gray2
dark:bg-dark-gray1
"""

main_content = "min-h-screen  p-4 relative bg-light-white dark:bg-dark-gray1 mx-auto flex flex-col gap-2 py-8 md:py-12 md:pb-8 lg:py-24 lg:pb-20 "

def with_theme_colors(class_name):
  return f"{class_name} text-light-gray1 hover:text-light-green dark:text-dark-page1 dark:hover:text-dark-gray2"

def with_theme_colors_text_no_hover(class_name):
  return f"{class_name} text-light-gray1 dark:text-dark-page1"


from .component import Component
from .page import Page
from typing import List, Dict, Any

class Layout:
    """
    Represents a layout in a Zenaura application.

    A Layout is a container for components, app routes. It is abstraction
    that handle the need for needing global components on app level such as nav bars, footers
    and others.
    """

    def __init__(self, top: List[Component], routes: List[Any], bottom: List[Component]):
        """
        Initializes a new Page instance.

        Args:
            top : list of component first children of root div
            routes: list of routes
            bottom: list of component bttom children after page rendering.
        """

        self.top = top
        self.routes = routes
        self.bottom = bottom

        if not isinstance(self.top, list):
            raise TypeError("top must be a list of components for root div first children")

        if not isinstance(self.bottom, list):
            raise TypeError("bottom must be a list of components for root div bottom children")

        if not isinstance(self.routes, dict):
            raise TypeError("routes should be a App.routes")

        for comp in self.top:
            if not isinstance(comp, Component):
                raise TypeError("every element in top must be instace of zenaura Component")
        
        for comp in self.bottom:
            if not isinstance(comp, Component):
                raise TypeError("every element in bottom must be instace of zenaura Component")
    



from .common import *


def Table(data, columns, attrs={}, class_names="min-w-full bg-white shadow-md rounded-xl", td_class_names="py-3 px-4 text-left", th_class_names="py-3 px-4", tr_class_names="border-b border-light-gray dark:border-dark-page1"):
	"""
	Creates a table component with given data, columns.
	args:
			data - list of dictionaries of data e.g. [{"key" : 1, "name": "Mike"}]
			columns - list of dictionaries of column names [{"title": "Name", "index": "name"}]
					note index for each column is where data is indexed and displayed, for example
							when index is "name" , for every dictionary in data "name" is fetched and displayed
			class_names - default class name
			attrs table tag attributes dictionary
	"""
	names = [Builder("th").with_class(th_class_names).with_text(col["title"]).build() for col in columns]
	indexes = [col["index"] for col in columns]
	rows = []
	for i in range(len(data)):
			row_items = [Builder("td").with_class(td_class_names).with_text(str(data[i][index])).build() for index in indexes]
			rows.append(Builder("tr").with_class(tr_class_names).with_children(
					*row_items
			).build())
	
	return    Builder("table").with_attributes(**attrs).with_attribute("class", class_names).with_children(
			Builder("thead").with_child(
					Builder("tr").with_class(tr_class_names).with_children(
					*names
			).build()
			).build(),
			Builder("tbody").with_children(
					*rows
			).build()
	).build()


import sys
import unittest
from unittest.mock import MagicMock, patch
from zenaura.client.page import Page
from zenaura.client.layout import Layout


sys.modules["pyscript"] = MagicMock()

class TestApp(unittest.TestCase):

    def setUp(self):  # Run before each test
        from zenaura.client.app import HistoryNode, PageHistory, NotFound, App, Route
        from .mocks.counter_mocks import Counter
        from .mocks.browser_mocks import MockDocument, MockWindow
        self.middleware = MagicMock()
        self.page = Page([Counter([])])
        self.route = Route("test", "/test",self.page, self.middleware  ) 
        self.new_route = Route
        self.router = App()
        self.router.add_route(self.route)

        self.history_node = HistoryNode
        self.page_history = PageHistory
        self.not_found = NotFound()
        self.document = MockDocument()
        self.window = MockWindow()
        self.dom_patcher = patch("zenaura.client.app.document", self.document)
        self.dom_patcher.start()
        self.window_patcher = patch("zenaura.client.app.window", self.window)
        self.window_patcher.start()
        self.page_with_empty_attrs = Page([Counter([])], {})
        self.page_with_attrs = Page([Counter([])], {"k" : "test"})

    def test_history_node_init(self):
        node = self.history_node()
        self.assertIsNone(node.page)
        self.assertIsNone(node.prev)
        self.assertIsNone(node.next)

    def test_page_history_visit(self):
        home_page = MagicMock()
        page_history = self.page_history()
        new_page = MagicMock()
        page_history.visit(new_page)
        self.assertEqual(page_history.current.page, new_page)

    def test_page_history_back(self):
        home_page = MagicMock()
        page_history = self.page_history()
        new_page = MagicMock()
        page_history.visit(home_page)
        page_history.visit(new_page)
        self.assertEqual(page_history.back(), home_page)

    def test_page_history_forward(self):
        home_page = MagicMock()
        page_history = self.page_history()
        new_page = MagicMock()
        page_history.visit(new_page)
        self.assertTrue(page_history.forward())

    def test_not_found_render(self):
        node = self.not_found.render()
        # Add assertions to check the structure of the generated node

    def test_route_init(self):
        route = self.route
        self.assertEqual(route.title, "test")
        self.assertEqual(route.path, "/test")
        # Add more assertions to check other attributes of the route

    def test_router_init(self):
        router = self.router
        self.assertEqual(router.paths, ["/test"])


    def test_page_with_attrs(self):
        # test passing attributes to page
        self.assertEqual(self.page_with_empty_attrs.attributes, {})
        self.assertTrue(self.page_with_attrs.attributes["k"])


class TestAsync(unittest.IsolatedAsyncioTestCase):
    def setUp(self):  # Run before each test
        from zenaura.client.app import HistoryNode, PageHistory, NotFound, App, Route
        from .mocks.counter_mocks import Counter
        from .mocks.browser_mocks import MockDocument, MockWindow
        self.middleware = MagicMock()
        self.page = Page([Counter([])])
        self.route = Route("test", "/test",self.page, self.middleware  ) 
        self.new_route = Route
        self.router = App()
        self.router.add_route(self.route)
        self.page1 = Page([Counter([])])
        self.page2 = Page([Counter([])])
        self.home_page = Page([Counter([])])
        self.history_node = HistoryNode
        self.page_history = PageHistory
        self.not_found = NotFound()
        self.document = MockDocument()
        self.window = MockWindow()
        self.dom_patcher = patch("zenaura.client.app.document", self.document)
        self.dom_patcher.start()
        self.window_patcher = patch("zenaura.client.app.window", self.window)
        self.window_patcher.start()
        self.page_with_empty_attrs = Page([Counter([])], {})
        self.page_with_attrs = Page([Counter([])], {"k" : "test"})
    
        self.router.history.current = self.history_node(self.page)
        self.document.setElementById(self.page.id, self.page)
        self.document.setElementById(self.page1.id, self.page1)
        self.document.setElementById(self.page2.id, self.page2)
        self.document.setElementById(self.home_page.id, self.home_page)

    # async def test_router_navigate_to_unknown_path(self):

    #     # Navigating to an unknown path should trigger the not-found route
    #     unknown_path = "/unknown"
    #     await self.router.navigate(unknown_path)
    #     self.assertEqual(self.document.title, "Page Not Found")
    #     # Add more assertions to check if the NotFound component is mounted and middleware is called

    # async def test_router_navigate_to_known_path(self):
    #     # Navigating to a known path should mount the associated page and update the document title
    #     known_path = "/test"
    #     await self.router.navigate(known_path)
    #     self.assertEqual(self.document.title, "test")
    #     # Add more assertions to check if the page is mounted and middleware is called

    # async def test_router_navigate_back(self):
    #     self.router.history.current = self.history_node(self.page)

    #     # Navigating back should return to the previous page in the history
    #     home_page = MagicMock()
    #     page1 = MagicMock()
    #     page2 = MagicMock()
    #     self.document.setElementById("page1", page1)
    #     self.document.setElementById("page2", page2)
    #     self.document.setElementById("homepage", home_page)
    #     page_history = self.page_history()
    #     page_history.visit(home_page)
    #     page_history.visit(page1)
    #     page_history.visit(page2)
    #     self.router.handle_location()  # Simulate navigating to page2
    #     await self.router.back()
    #     # Add assertions to check if page1 is mounted and document title is updated
        
    # async def test_router_navigate_forward(self):
    #     # Navigating forward should move to the next page in the history
    #     self.router.history.current = self.history_node(self.page)

    #     page_history = self.page_history()
    #     page_history.visit(self.home_page)
    #     page_history.visit(self.page1)
    #     page_history.visit(self.page2)
    #     self.router.handle_location()  # Simulate navigating to page2
    #     print("ids", self.document.elementsById)
    #     await self.router.back()
    #     # await self.router.forward()


    # async def test_router_wildcard_route_matching(self):
    #     route = self.new_route("wildcard", "/users/*", self.page , None)
    #     self.router.add_route(route)
    #     self.router.history.current = self.history_node(self.page)
    #     await self.router.navigate("/users/123/123?k=1&k2=3")
    #     self.window.location.pathname = "/users/123/123?k=1&k2=3"
    #     current_route, info = self.router.get_current_route()
    #     self.assertEqual(info["wildcard"]["params"], ["123", "123"])
    #     self.assertEqual(info["wildcard"]["query"]["k"], "1" )
    #     self.assertEqual(info["wildcard"]["query"]["k2"], "3" )


    # async def test_router_nested_routes(self):
    #     # Test navigation with nested routes
    #     parent_route = self.new_route("parent", "/parent", self.page, None)
    #     child_route = self.new_route("child", "/parent/child", self.page, None)

    #     self.router.add_route(parent_route)
    #     self.router.add_route(child_route)
    #     await self.router.navigate("/parent/child")
    #     self.window.location.pathname = "/parent/child"
    #     current_route, info = self.router.get_current_route()
    #     self.assertEqual(current_route[0].id, child_route.page.id)

    # async def test_router_case_sensitive_paths(self):
    #     # Test case sensitivity of route paths
    #     case_sensitive_route = self.new_route("case_sensitive", "/CaseSensitive", self.page, None)
    #     self.router.add_route(case_sensitive_route)
    #     await self.router.navigate("/casesensitive")  # Attempting to navigate with different case
    #     self.assertEqual(self.document.title, "Page Not Found")

    # async def test_router_middleware_execution_order(self):
    #     # Test the execution order of middleware
    #     middleware_order = []

    #     def middleware1():
    #         middleware_order.append(1)

    #     def middleware2():
    #         middleware_order.append(2)

    #     def middleware():
    #         middleware1()
    #         middleware2()

    #     route_with_middleware = self.new_route("middleware_order", "/middleware", self.page, middleware)
    #     self.router.add_route(route_with_middleware)
    #     await self.router.navigate("/middleware")
    #     self.assertEqual(middleware_order, [1, 2])

    async def test_layout_attached(self):
        from zenaura.client.component import Component, Reuseable
        magic_method = MagicMock() 

        class TestComponentMagicMethodAttached(Component):
            def __init__(self, mock_method):
                self.mock_method = mock_method

            async def attached(self):
                self.mock_method()
        attached_comp = TestComponentMagicMethodAttached(magic_method)
        app_layout = Layout(
            [attached_comp],
            self.router.routes, 
            [attached_comp]
        )
        self.router.layout = app_layout
        await self.router.navigate("/test")
        self.assertEqual(attached_comp.mock_method.call_count, 2)


"""
Pyodide and pyscript is intended to work in the browser. 
This file have conditional imports so it works magically
on server and client.

On the server pyodide, pyscript will be missing.

Users of zenaura don't need to know about this,

they just import from this file 
e.g. 
from zenaura.web.utils import document 

This is used across the source code as well.

"""
in_browser = False

global document
global window
global create_proxy 
global to_js

try:
    from pyscript import document as pyscript_doc, window as pyscript_window
    in_browser = True 
    document = pyscript_doc
    window = pyscript_window
except:
    from zenaura.mocks.browser_mocks import MockDocument, MockWindow
    document = MockDocument()
    window = MockWindow()
    

try : 
     from pyodide.ffi import create_proxy_pyodide, to_js_poyodide
     create_proxy = create_proxy_pyodide
     to_js = to_js_poyodide
except:
    # TODO: need to create mocks
    create_proxy = lambda x: x 
    to_js = lambda x: x 



from .common import *


def TabButton(tab_number, label, active_tab_variable, on_click, class_name, active_class=" me-2 inline-block p-5 border-b-2 border-light-green border-spacing-2 dark:border-dark-page1"):
	class_expression = active_class if tab_number == active_tab_variable else ""
	return ButtonWithAttrsChildren(
			class_name=class_name  + class_expression,
			attrs={
					"py-click": on_click,
					"name": tab_number
			},
			children=[label]
	)

def TabContent(tab_number, active_tab_variable, content):
	is_visible = "block" if tab_number == active_tab_variable else "hidden"
	return Div(
			f"{is_visible} transition-all duration-300 p-4",
			[
				content    
			]
	)

def Tabs(buttons, content, g_class_names ="bg-gray-100 font-sans", buttons_wrapper_class_names="flex  border-b-2 border-light-green dark:border-dark-black"):
	"""
			Create tabs with content for each tab. 
			args:
					buttons : list of TabButton which is a button upon click becomes active and display content
					content : list of TabContent which is the content under each tab.
					g_class_names: global wrapper div class names
					button_wrapper_class_names: button wrapper div class names.
	"""
	return Div(
	g_class_names,
	[
	Div(
		"p-8",
		[
		Div(
			"",
			[
				Div(
					buttons_wrapper_class_names,
					buttons
				),
				Div(
					"mt-3", 
					content
				)
			]
		)
		]
	)
	]
)


from setuptools import setup, find_packages

with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='zenaura',
    version='0.15.2',
    description="Zenaura is an experimental Python library built upon PyScript, designed to empower Python developers to create stateful, component-based Single Page Applications (SPAs). By leveraging a virtual DOM implementation, Zenaura optimizes the performance, reactivity, responsiveness, and interactivity of web applications. This allows developers to build high-performance, dynamic web applications using familiar Python concepts and syntax.",
    author="Ahmed Rakan",
    author_email="ar.aldhafeeri11@gmail.com",
    packages=[
        'zenaura', 
        'zenaura.server', 
        'zenaura.client', 
        'zenaura.client.algorithm', 
        'zenaura.client.compiler', 
        'zenaura.client.hydrator', 
        'zenaura.client.observer', 
        'zenaura.client.tags', 
        'zenaura.client.dom', 
        'zenaura.client.dom.lifecycles', 
        'zenaura.cli', 
        'zenaura.ui', 
        "zenaura.web",
        "zenaura.mocks"

        ],
    install_requires=[
        'bleach==6.1.0', 
        'nest_asyncio==1.6.0'
    ],
    entry_points={
        "console_scripts": [
            "zenaura = zenaura.cli:main",
        ],
    },
    test_suite='tests',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
)


from .common import Span, LI, A, OL
from zenaura.client.tags.builder import Builder


def BreadCrumbs(
    breadcrumbs, 
    seprator=">", 
    span_class="", 
    a_tag_class="opacity-60", 
    li_class="flex cursor-pointer items-center font-sans text-sm font-normal leading-normal antialiased transition-colors duration-300 text-light-gray1 hover:text-light-green dark:text-dark-page1 dark:hover:text-dark-gray2 ", 
    ol_class="flex w-full flex-wrap items-center rounded-md py-2 px-4", 
    sep_class="pointer-events-none mx-2 select-none font-sans text-sm font-normal leading-normal antialiased"
    ) -> "BreadCrumbs":
  """
    Display the path to the current resources as hierarchy of links

    args : 
      breadcrumbs - list of tuples as [("title", "handler"), ...] displayed in order
        handler url to navigate to.
        title is the title of breadcrumb that will be displayed.
      seprator - custom seprator between breadcrumbs default is text >
      span_class - span element classes
      a_tag_class - breadcrumb a tag css calsses
      li_class - breadcrumb li tag css classes
      ol_class - bread crumb ol tag css classes
      sep_class - breadcrumb seprator css classes
  """
  crumbs = []
  sep = Span(sep_class, seprator)
  n = len(breadcrumbs)
  for i in range(n):
    title, handler  = breadcrumbs[i]
    crumb = LI(
      A(
        Span(span_class, title),
        {"class": a_tag_class, "href": handler},
      ),
      {"class": li_class}
    )
    crumbs.append(crumb)
    if i != n - 1:
      crumbs.append(sep)
  return OL(crumbs, {"class" : ol_class})



from zenaura.client.tags.builder import Builder

def Badge(
    text,
    attrs, 
    default_class="p-2 text-xs font-medium me-2 px-2.5 py-0.5 rounded bg-light-green text-light-gray1 dark:text-dark-page1 dark:bg-dark-gray2", 
    ):
    """
      Displays a form input field. 
      args:
        text : badge text.
        attrs: span attributes.
        default_class: default css class names.
      
    """
    return Builder("span").with_text(text).with_attribute("class", default_class).with_attributes(**attrs).build()